#include <msp/gl/renderer.h>
#include "stereocombiner.h"
#include "stereoview.h"

using namespace std;

namespace Msp {
namespace VR {

StereoView::StereoView(unsigned w, unsigned h, const GL::Camera &c, const GL::Renderable &r, const StereoCombiner &m):
	width(w),
	height(h),
	base_camera(c),
	renderable(r),
	combiner(0)
{
	set_combiner(m);
	set_eye_spacing(0.07);
}

void StereoView::set_combiner(const StereoCombiner &c)
{
	combiner = &c;

	unsigned w = width/combiner->get_width_divisor()*combiner->get_oversize();
	unsigned h = height/combiner->get_height_divisor()*combiner->get_oversize();
	left.create_target(w, h);
	right.create_target(w, h);
}

void StereoView::set_eye_spacing(float s)
{
	eye_spacing = s;
}

void StereoView::setup_frame() const
{
	offset_axis = normalize(cross(base_camera.get_look_direction(), base_camera.get_up_direction()))*0.5f;

	EyeParams params;
	params.fov = combiner->get_field_of_view();
	if(params.fov==Geometry::Angle<float>::zero())
		params.fov = base_camera.get_field_of_view();

	params.aspect = base_camera.get_aspect();
	if(!combiner->is_aspect_kept())
		params.aspect = params.aspect*combiner->get_height_divisor()/combiner->get_width_divisor();

	params.near_clip = base_camera.get_near_clip();
	params.far_clip = base_camera.get_far_clip();

	left.setup_frame(base_camera, offset_axis*-eye_spacing, params);
	right.setup_frame(base_camera, offset_axis*eye_spacing, params);

	renderable.setup_frame();
}

void StereoView::finish_frame() const
{
	renderable.finish_frame();
}

void StereoView::render(const GL::Tag &tag) const
{
	setup_frame();
	left.render(renderable, tag);
	right.render(renderable, tag);
	combiner->render(left.target->color, right.target->color);
	finish_frame();
}

void StereoView::render(GL::Renderer &renderer, const GL::Tag &tag) const
{
	renderer.escape();
	return render(tag);
}


StereoView::RenderTarget::RenderTarget(unsigned width, unsigned height)
{
	color.set_min_filter(GL::LINEAR);
	color.set_wrap(GL::CLAMP_TO_EDGE);
	color.storage(GL::RGB, width, height);
	fbo.attach(GL::COLOR_ATTACHMENT0, color);

	depth.storage(GL::DEPTH_COMPONENT, width, height);
	fbo.attach(GL::DEPTH_ATTACHMENT, depth);
}


StereoView::Eye::Eye():
	target(0)
{ }

void StereoView::Eye::create_target(unsigned w, unsigned h)
{
	delete target;
	target = new RenderTarget(w, h);
}

void StereoView::Eye::setup_frame(const GL::Camera &base_camera, const GL::Vector3 &offset, const EyeParams &params) const
{
	camera.set_position(base_camera.get_position()+offset);
	camera.set_up_direction(base_camera.get_up_direction());
	camera.set_look_direction(base_camera.get_look_direction());

	camera.set_field_of_view(params.fov);
	camera.set_aspect(params.aspect);
	camera.set_depth_clip(params.near_clip, params.far_clip);
}

void StereoView::Eye::render(const GL::Renderable &renderable, const GL::Tag &tag) const
{
	GL::Bind bind_fbo(target->fbo);
	target->fbo.clear(GL::COLOR_BUFFER_BIT|GL::DEPTH_BUFFER_BIT);
	GL::Renderer renderer(&camera);
	renderable.render(renderer, tag);
}

} // namespace VR
} // namespace Msp
